/*
 * Copyright (C) 2025 Vasiliy Doylov <nekocwd@mainlining.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * Author: Vasiliy Doylov <nekocwd@mainlining.org>
 */

using Phosh;
using GLib;

[GtkTemplate (ui = "/mobi/phosh/plugins/media-player/player.ui")]
public class MediaPlayer.Player : Gtk.Box {
  [GtkChild]
  private unowned Gtk.Image image;
  [GtkChild]
  private unowned Gtk.Label title;
  [GtkChild]
  private unowned Gtk.Label artist;
  [GtkChild]
  private unowned Gtk.Button pause;
  [GtkChild]
  private unowned Gtk.Button back;
  [GtkChild]
  private unowned Gtk.Button next;
  [GtkChild]
  private unowned Gtk.Image pause_img;
  [GtkChild]
  private unowned Gtk.ProgressBar progress;

  delegate void OnStyleGenerated (Gdk.RGBA fg, Gdk.RGBA start, Gdk.RGBA end);

  private string id = Uuid.string_random ();
  private Gtk.CssProvider css_provider = new Gtk.CssProvider ();
  construct {
    info ("Constructor");
    Gtk.StyleContext.add_provider_for_screen
    (
     Gdk.Screen.get_default (),
     css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION);
    get_style_context ().add_class (id);
  }

  ~Player () {
    info ("Destructor");
    Gtk.StyleContext.remove_provider_for_screen (Gdk.Screen.get_default (), css_provider);
  }

  void set_colors (Gdk.RGBA start, Gdk.RGBA end, Gdk.RGBA fg) {
    var progress = fg.copy ();
    progress.alpha = 0.5;
    var hover = fg.copy ();
    hover.alpha = 0.2;
    var data =
      @"
      .$id {
        color: $fg;
        background: radial-gradient(circle at left, $start 0%, $end 100%);
      }
      .$id .nekoplayer-controls progressbar > trough {
        background-color: $progress;
      }
      .$id .nekoplayer-controls progressbar > trough > progress {
        background-color: $fg;
      }
      .$id .nekoplayer-controls button:hover {
        background-color: $hover;
      }
      ";
    try {
      css_provider.load_from_data (data);
    } catch (Error err) {
      warning ("Something wrong with CSS generation: %s", err.message);
    }
  }

  private void generate_style (Gdk.Pixbuf pixbuff) {
    var histogram = Color.quanitize_pixbuff (pixbuff);
    Gdk.RGBA? fg_color = null, bg_start_color = null, bg_end_color = null;
    double bg_start_lum = 0;

    bg_start_color = Color.get_most_common_color (ref histogram);
    bg_start_lum = Color.get_pixel_luminance (bg_start_color);
    histogram.unset (bg_start_color);
    bg_end_color = Color.get_most_common_color (ref histogram, bg_start_lum, 0.3);
    histogram.unset (bg_end_color);
    fg_color = Color.get_most_common_color (ref histogram, bg_start_lum, 0.5, true);

    if (bg_end_color == null) {
      message ("Failed to chose end background color color. Fallbacking to start color");
      bg_end_color = bg_start_color;
    }
    if (fg_color == null) {
      message ("Failed to chose forground color. Fallbacking to contrast");
      if (bg_start_lum > 0.5)
        fg_color = Gdk.RGBA () {
          red = 0,
          green = 0,
          blue = 0,
          alpha = 1,
        };
      else
        fg_color = Gdk.RGBA () {
          red = 1,
          green = 1,
          blue = 1,
          alpha = 1,
        };
    }

    Idle.add (() => {
      set_colors (bg_start_color, bg_end_color, fg_color);
      return false;
    });
  }

  private void run_color_thred (Gdk.Pixbuf pb) {
    new Thread<void> ("Picture", () => generate_style (pb));
  }

  private Icon ? process_icon (Icon? icon) {
    Gdk.Pixbuf pixbuff = null;
    if (icon == null)
      return icon;

    if (icon is Gdk.Pixbuf)
      pixbuff = (Gdk.Pixbuf) icon;
    else if (icon is FileIcon) {
      var fi = (LoadableIcon) icon;
      try {
        pixbuff = new Gdk.Pixbuf.from_stream (fi.load (1024, null, null));
      } catch (Error err) {
        warning ("Icon load failed %s", err.message);
        return icon;
      }
    } else {
      return icon;
    }

    // Start background generator
    run_color_thred (pixbuff.scale_simple (64, 64, Gdk.InterpType.BILINEAR));

    // TODO: REMOVE WHEN PHOSH MOVES TO GTK4
    pixbuff = Utils.round_pixbuff (pixbuff, (110.0 - 10.0 /* pixel-size - padding */) / 10 / 100);
    return Utils.rect_pixbuff (pixbuff);
  }

  private Binding[] bindings = new Binding[0];
  private Utils.SignalHelp[] signals = new Utils.SignalHelp[0];
  public void bind_to_player (Gtk.Box player) {
    player.bind_property ("visible", this, "visible", BindingFlags.SYNC_CREATE);
    // Buttons
    var play_btn = (Gtk.Button) Utils.find_widget_by_css_name (player, "btn_play");
    signals += Utils.SignalHelp () {
      object = pause,
      signal = pause.clicked.connect (() => play_btn.clicked ())
    };
    var play_img = (Gtk.Image) Utils.find_widget_by_css_name (player, "img_play");
    bindings += play_img.bind_property ("icon-name", pause_img, "icon-name", BindingFlags.SYNC_CREATE);
    var prev_btn = (Gtk.Button) Utils.find_widget_by_css_name (player, "btn_prev");
    signals += Utils.SignalHelp () {
      object = back,
      signal = back.clicked.connect (() => prev_btn.clicked ())
    };
    var next_btn = (Gtk.Button) Utils.find_widget_by_css_name (player, "btn_next");
    signals += Utils.SignalHelp () {
      object = next,
      signal = next.clicked.connect (() => next_btn.clicked ())
    };
    // Labels
    var _title = (Gtk.Label) Utils.find_widget_by_css_name (player, "lbl_title");
    bindings += _title.bind_property ("label", title, "label", BindingFlags.SYNC_CREATE);
    var _artist = (Gtk.Label) Utils.find_widget_by_css_name (player, "lbl_artist");
    bindings += _artist.bind_property ("label", artist, "label", BindingFlags.SYNC_CREATE);
    // Image
    var _image = (Gtk.Image) Utils.find_widget_by_css_name (player, "img_art");
    bindings += _image.bind_property ("gicon", image, "gicon", BindingFlags.SYNC_CREATE, (binding, src, ref dst) => {
      dst = process_icon ((Icon) src);
      return true;
    });
    // Progress bar
    var _progress = (Gtk.ProgressBar) Utils.find_widget_by_css_name (player, "prb_position");
    bindings += _progress.bind_property ("fraction", progress, "fraction", BindingFlags.SYNC_CREATE);
  }

  public void unbind () {
    foreach (var b in bindings) {
      b.unbind ();
    }
    foreach (var s in signals) {
      s.object.disconnect (s.signal);
    }
  }

  public void bind_settings (string key) {
    var settings = new Settings.with_path (Preferences.Keys.SCHEMA_ID, Preferences.Keys.SETTINGS_PATH);
    settings = settings.get_child (key);
    /*
     * Margins
     */
    settings.bind (Preferences.Keys.MARGIN_TOP, this, Preferences.Keys.MARGIN_TOP, SettingsBindFlags.DEFAULT);
    settings.bind (Preferences.Keys.MARGIN_BOTTOM, this, Preferences.Keys.MARGIN_BOTTOM, SettingsBindFlags.DEFAULT);
    settings.bind (Preferences.Keys.MARGIN_START, this, Preferences.Keys.MARGIN_START, SettingsBindFlags.DEFAULT);
    settings.bind (Preferences.Keys.MARGIN_END, this, Preferences.Keys.MARGIN_END, SettingsBindFlags.DEFAULT);
  }
}
